from __future__ import annotations

from http import HTTPStatus

import pytest
from django.test import override_settings
from django.urls import reverse

from django_bird.staticfiles import AssetType

from .utils import TestAsset
from .utils import TestComponent


@pytest.fixture(autouse=True)
def debug_mode():
    with override_settings(DEBUG=True):
        yield


def test_url_reverse(templates_dir):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )
    button_css = TestAsset(
        component=button,
        content=".button { color: blue; }",
        asset_type=AssetType.CSS,
    ).create()

    assert reverse(
        "django_bird:asset",
        kwargs={"component_name": button.name, "asset_filename": button_css.file.name},
    )


def test_get_css(templates_dir, client):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )
    button_css = TestAsset(
        component=button,
        content=".button { color: blue; }",
        asset_type=AssetType.CSS,
    ).create()

    response = client.get(f"/__bird__/assets/{button.name}/{button_css.file.name}")

    assert response.status_code == HTTPStatus.OK
    assert response["Content-Type"] == "text/css"

    content = b"".join(response.streaming_content).decode()

    assert content == ".button { color: blue; }"


def test_get_js(templates_dir, client):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )
    button_js = TestAsset(
        component=button, content="console.log('button');", asset_type=AssetType.JS
    ).create()

    response = client.get(f"/__bird__/assets/{button.name}/{button_js.file.name}")

    assert response.status_code == HTTPStatus.OK
    assert response["Content-Type"] == "application/javascript"

    content = b"".join(response.streaming_content).decode()

    assert content == "console.log('button');"


def test_get_invalid_type(templates_dir, client):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )

    button_txt_file = button.file.parent / f"{button.file.stem}.txt"
    button_txt_file.write_text("hello from a text file")

    response = client.get(f"/__bird__/assets/{button.name}/{button_txt_file.name}")

    assert response.status_code == HTTPStatus.NOT_FOUND


def test_get_nonexistent_component(templates_dir, client):
    bird_dir = templates_dir / "bird"
    bird_dir.mkdir(exist_ok=True)

    nonexistent_css_file = bird_dir / "nonexistent.css"
    nonexistent_css_file.write_text(".button { color: blue; }")

    response = client.get(f"/__bird__/assets/nonexistent/{nonexistent_css_file.name}")

    assert response.status_code == HTTPStatus.NOT_FOUND


def test_asset_view_nonexistent_asset(templates_dir, client):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )

    response = client.get(f"/__bird__/assets/{button.name}/button.css")

    assert response.status_code == HTTPStatus.NOT_FOUND


def test_asset_view_warns_debug_false(templates_dir, client):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )
    button_css = TestAsset(
        component=button,
        content=".button { color: blue; }",
        asset_type=AssetType.CSS,
    ).create()

    with override_settings(DEBUG=False):
        with pytest.warns(RuntimeWarning):
            response = client.get(
                f"/__bird__/assets/{button.name}/{button_css.file.name}"
            )

    assert response.status_code == HTTPStatus.OK
    assert response["Content-Type"] == "text/css"



from __future__ import annotations

from django import template

from .tags import asset
from .tags import bird
from .tags import prop
from .tags import slot

register = template.Library()


register.tag(asset.CSS_TAG, asset.do_asset)
register.tag(asset.JS_TAG, asset.do_asset)
register.tag(bird.TAG, bird.do_bird)
register.tag(prop.TAG, prop.do_prop)
register.tag(slot.TAG, slot.do_slot)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any

from django_bird.staticfiles import AssetType


@dataclass
class TestComponent:
    __test__ = False

    name: str
    content: str
    file: Path | None = None
    parent_dir: str = "bird"
    sub_dir: str | None = None

    def create(self, base_dir: Path) -> TestComponent:
        parent = base_dir / self.parent_dir
        parent.mkdir(exist_ok=True)

        if self.sub_dir is not None:
            dir = parent / self.sub_dir
            dir.mkdir(exist_ok=True)
        else:
            dir = parent

        template = dir / f"{self.name}.html"
        template.write_text(self.content)

        self.file = template

        return self


@dataclass
class TestAsset:
    __test__ = False

    component: TestComponent
    content: str
    asset_type: AssetType
    file: Path | None = None

    def create(self) -> TestAsset:
        if self.component.file is None:
            raise ValueError("Component must be created before adding assets")

        component_dir = self.component.file.parent
        component_name = self.component.file.stem

        asset_file = component_dir / f"{component_name}{self.asset_type.ext}"
        asset_file.write_text(self.content)

        self.file = asset_file

        return self


@dataclass
class TestComponentCase:
    __test__ = False

    component: TestComponent
    template_content: str
    expected: str
    description: str = ""
    template_context: dict[str, Any] = field(default_factory=dict)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from hashlib import md5
from pathlib import Path

from cachetools import LRUCache
from django.apps import apps
from django.conf import settings
from django.template.backends.django import Template as DjangoTemplate
from django.template.context import Context
from django.template.engine import Engine
from django.template.exceptions import TemplateDoesNotExist
from django.template.loader import select_template

from django_bird.staticfiles import Asset

from .conf import app_settings
from .staticfiles import AssetType
from .templates import get_template_names


@dataclass(frozen=True, slots=True)
class Component:
    name: str
    template: DjangoTemplate
    assets: frozenset[Asset] = field(default_factory=frozenset)

    def get_asset(self, asset_filename: str) -> Asset | None:
        for asset in self.assets:
            if asset.path.name == asset_filename:
                return asset
        return None

    def render(self, context: Context):
        return self.template.template.render(context)

    @property
    def id(self):
        normalized_source = "".join(self.source.split())
        hashed = md5(
            f"{self.name}:{self.path}:{normalized_source}".encode()
        ).hexdigest()
        return hashed[:7]

    @property
    def nodelist(self):
        return self.template.template.nodelist

    @property
    def path(self):
        return self.template.template.origin.name

    @property
    def source(self):
        return self.template.template.source

    @classmethod
    def from_name(cls, name: str):
        template_names = get_template_names(name)
        template = select_template(template_names)

        assets: set[Asset] = set()
        template_path = Path(template.template.origin.name)
        for asset_type in AssetType:
            asset_path = template_path.with_suffix(asset_type.ext)
            if asset_path.exists():
                asset = Asset.from_path(asset_path)
                assets.add(asset)

        return cls(name=name, template=template, assets=frozenset(assets))


class ComponentRegistry:
    def __init__(self, maxsize: int = 100):
        self._components: LRUCache[str, Component] = LRUCache(maxsize=maxsize)

    def get_dirs(self):
        engine = Engine.get_default()
        dirs: list[str | Path] = list(engine.dirs)

        for app_config in apps.get_app_configs():
            template_dir = Path(app_config.path) / "templates"
            if template_dir.is_dir():
                dirs.append(template_dir)

        base_dir = getattr(settings, "BASE_DIR", None)

        if base_dir is not None:
            root_template_dir = Path(base_dir) / "templates"
            if root_template_dir.is_dir():
                dirs.append(root_template_dir)

        return dirs

    def discover_components(self) -> None:
        template_dirs = self.get_dirs()

        component_dirs = [
            Path(template_dir) / component_dir
            for template_dir in template_dirs
            for component_dir in app_settings.COMPONENT_DIRS + ["bird"]
        ]

        for component_dir in component_dirs:
            component_dir = Path(component_dir)

            if not component_dir.is_dir():
                continue

            for component_path in component_dir.rglob("*.html"):
                component_name = str(
                    component_path.relative_to(component_dir).with_suffix("")
                ).replace("/", ".")
                try:
                    component = Component.from_name(component_name)
                    self._components[component_name] = component
                except TemplateDoesNotExist:
                    continue

    def clear(self) -> None:
        """Clear the registry. Mainly useful for testing."""
        self._components.clear()

    def get_component(self, name: str) -> Component:
        try:
            if not settings.DEBUG:
                return self._components[name]
        except KeyError:
            pass

        component = Component.from_name(name)
        self._components[name] = component
        return component

    def get_assets(self, asset_type: AssetType) -> frozenset[Asset]:
        assets: set[Asset] = set()
        for component in self._components.values():
            for asset in component.assets:
                if asset.type == asset_type:
                    assets.add(asset)
        return frozenset(assets)


components = ComponentRegistry()



from __future__ import annotations

import warnings
from io import BytesIO

from django.conf import settings
from django.http import FileResponse
from django.http import Http404
from django.http.request import HttpRequest
from django.template.exceptions import TemplateDoesNotExist

from .components import components


def asset_view(request: HttpRequest, component_name: str, asset_filename: str):
    if not settings.DEBUG:
        warnings.warn(
            "Serving assets through this view in production is not recommended.",
            RuntimeWarning,
            stacklevel=2,
        )

    try:
        component = components.get_component(component_name)
    except (KeyError, TemplateDoesNotExist) as err:
        raise Http404("Component not found") from err

    asset = component.get_asset(asset_filename)
    if not asset or not asset.path.exists():
        raise Http404("Asset not found")

    content = asset.path.read_bytes()

    return FileResponse(BytesIO(content), content_type=asset.type.content_type)



# pyright: reportAny=false
from __future__ import annotations

from django import template
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import TagBits
from django_bird._typing import override

TAG = "bird:prop"


def do_prop(parser: Parser, token: Token) -> PropNode:
    bits = token.split_contents()
    name, default = parse_prop_name(bits)
    attrs = parse_attrs(bits)
    return PropNode(name, default, attrs)


def parse_prop_name(bits: TagBits) -> tuple[str, str | None]:
    if len(bits) <= 1:
        msg = f"{TAG} tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    try:
        name, default = bits[1].split("=")
        return name, default.strip("'\"")
    except ValueError:
        return bits[1], None


def parse_attrs(bits: TagBits) -> TagBits:
    return bits[2:]


class PropNode(template.Node):
    def __init__(self, name: str, default: str | None, attrs: TagBits):
        self.name = name
        self.default = default
        self.attrs = attrs

    @override
    def render(self, context: Context) -> str:
        return ""



from __future__ import annotations

import contextlib
import logging
import re
from pathlib import Path

import pytest
from django.conf import settings
from django.template.backends.django import DjangoTemplates
from django.template.backends.django import Template as DjangoTemplate
from django.template.engine import Engine
from django.test import override_settings
from django.urls import clear_url_caches
from django.urls import include
from django.urls import path

from .settings import DEFAULT_SETTINGS

pytest_plugins = []


def pytest_configure(config):
    logging.disable(logging.CRITICAL)

    settings.configure(**DEFAULT_SETTINGS, **TEST_SETTINGS)


TEST_SETTINGS = {
    "INSTALLED_APPS": [
        "django_bird",
    ],
    "TEMPLATES": [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": [
                Path(__file__).parent / "templates",
            ],
            "OPTIONS": {
                "builtins": [
                    "django.template.defaulttags",
                    "django_bird.templatetags.django_bird",
                ],
                "loaders": [
                    "django_bird.loader.BirdLoader",
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ],
            },
        }
    ],
}


@pytest.fixture(autouse=True)
def setup_urls():
    urlpatterns = [
        path("__bird__/", include("django_bird.urls")),
    ]

    clear_url_caches()

    with override_settings(
        ROOT_URLCONF=type(
            "urls",
            (),
            {"urlpatterns": urlpatterns},
        ),
    ):
        yield

    clear_url_caches()


@pytest.fixture
def templates_dir(tmp_path):
    templates_dir = tmp_path / "templates"
    templates_dir.mkdir()
    return templates_dir


@pytest.fixture(autouse=True)
def override_templates_settings(templates_dir):
    with override_settings(
        TEMPLATES=[
            settings.TEMPLATES[0]
            | {
                "DIRS": [
                    *settings.TEMPLATES[0]["DIRS"],
                    templates_dir,
                ]
            }
        ]
    ):
        yield


@pytest.fixture
def override_app_settings():
    from django_bird.conf import DJANGO_BIRD_SETTINGS_NAME

    @contextlib.contextmanager
    def _override_app_settings(**kwargs):
        with override_settings(**{DJANGO_BIRD_SETTINGS_NAME: {**kwargs}}):
            yield

    return _override_app_settings


@pytest.fixture(autouse=True)
def data_bird_attr_app_setting(override_app_settings, request):
    from django_bird.conf import app_settings

    enable = (
        app_settings.ENABLE_BIRD_ID_ATTR
        if "default_app_settings" in request.keywords
        else False
    )

    with override_app_settings(ENABLE_BIRD_ID_ATTR=enable):
        yield


@pytest.fixture
def create_template():
    def _create_template(template_file: Path) -> DjangoTemplate:
        engine = Engine(
            builtins=["django_bird.templatetags.django_bird"],
            dirs=[str(template_file.parent)],
            loaders=["django_bird.loader.BirdLoader"],
        )
        template = engine.get_template(template_file.name)
        backend = DjangoTemplates(
            {
                "NAME": "django",
                "DIRS": [],
                "APP_DIRS": True,
                "OPTIONS": {
                    "autoescape": True,
                    "debug": False,
                    "context_processors": [],
                },
            }
        )
        return DjangoTemplate(template, backend)

    return _create_template


@pytest.fixture
def normalize_whitespace():
    def func(text: str) -> str:
        """Normalize whitespace in rendered template output"""
        # multiple whitespace characters
        text = re.sub(r"\s+", " ", text)
        # after opening tag, including when there are attributes
        text = re.sub(r"<(\w+)(\s+[^>]*)?\s*>", r"<\1\2>", text)
        # before closing tag
        text = re.sub(r"\s+>", ">", text)
        # after opening tag and before closing tag
        text = re.sub(r">\s+<", "><", text)
        # immediately after opening tag (including attributes) or before closing tag
        text = re.sub(r"(<\w+(?:\s+[^>]*)?>)\s+|\s+(<\/\w+>)", r"\1\2", text)
        # between tags and text content
        text = re.sub(r">\s+([^<])", r">\1", text)
        text = re.sub(r"([^>])\s+<", r"\1<", text)
        return text.strip()

    return func


@pytest.fixture(autouse=True)
def clear_components_registry():
    from django_bird.components import components

    components.clear()
    yield
    components.clear()



from __future__ import annotations

import hashlib
import re

from django.core.cache import cache
from django.template.base import Node
from django.template.base import Origin
from django.template.base import Template
from django.template.context import Context
from django.template.engine import Engine
from django.template.loader_tags import ExtendsNode
from django.template.loader_tags import IncludeNode
from django.template.loaders.filesystem import Loader as FileSystemLoader

from ._typing import _has_nodelist
from ._typing import override
from .compiler import Compiler
from .components import components
from .templatetags.tags.bird import TAG
from .templatetags.tags.bird import BirdNode

BIRD_TAG_PATTERN = re.compile(rf"{{%\s*{TAG}\s+([^\s%}}]+)")


class BirdLoader(FileSystemLoader):
    def __init__(self, engine: Engine):
        super().__init__(engine)
        self.compiler = Compiler()

    @override
    def get_contents(self, origin: Origin) -> str:
        contents = super().get_contents(origin)

        if not BIRD_TAG_PATTERN.search(contents):
            return contents

        template = Template(contents, origin=origin, engine=self.engine)
        context = Context()
        with context.bind_template(template):
            self._ensure_components_loaded(template, context)

        cache_key = f"bird_component_{hashlib.md5(contents.encode()).hexdigest()}"
        compiled = cache.get(cache_key)
        if compiled is None:
            compiled = self.compiler.compile(contents)
            cache.set(cache_key, compiled, timeout=None)
        return compiled

    def _ensure_components_loaded(
        self, node: Template | Node, context: Context
    ) -> None:
        """Ensure all components used in the template are loaded."""
        if isinstance(node, BirdNode):
            components.get_component(node.name)

        if not _has_nodelist(node) or node.nodelist is None:
            return

        for child in node.nodelist:
            if isinstance(child, BirdNode):
                components.get_component(child.name)

            if isinstance(child, ExtendsNode):
                parent_template = child.get_parent(context)
                self._ensure_components_loaded(parent_template, context)

            if isinstance(child, IncludeNode):
                template_name = child.template.token.strip("'\"")
                included_template = self.engine.get_template(template_name)
                self._ensure_components_loaded(included_template, context)

            if hasattr(child, "nodelist"):
                self._ensure_components_loaded(child, context)



# pyright: reportAny=false
from __future__ import annotations

from typing import cast

from django import template
from django.template.base import NodeList
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from django_bird._typing import TagBits
from django_bird._typing import override

TAG = "bird:slot"
END_TAG = "endbird:slot"


def do_slot(parser: Parser, token: Token) -> SlotNode:
    bits = token.split_contents()
    name = parse_slot_name(bits)
    nodelist = parse_nodelist(parser)
    return SlotNode(name, nodelist)


def parse_slot_name(bits: TagBits) -> str:
    if len(bits) == 1:
        return "default"
    elif len(bits) == 2:
        name = bits[1]
        if name.startswith("name="):
            name = name.split("=")[1]
        else:
            name = name
        return name.strip("'\"")
    else:
        msg = f"{TAG} tag requires either no arguments, one argument, or 'name=\"slot_name\"'"
        raise template.TemplateSyntaxError(msg)


def parse_nodelist(parser: Parser) -> NodeList:
    nodelist = parser.parse((END_TAG,))
    parser.delete_first_token()
    return nodelist


class SlotNode(template.Node):
    def __init__(self, name: str, nodelist: NodeList):
        self.name = name
        self.nodelist = nodelist

    @override
    def render(self, context: Context) -> SafeString:
        default_content: str = self.nodelist.render(context)
        slots = context.get("slots")

        if not slots or not isinstance(slots, dict):
            return mark_safe(default_content)

        slots_dict = cast(dict[str, str], slots)
        slot_content = slots_dict.get(self.name)

        if slot_content is None or slot_content == "":
            return mark_safe(default_content)

        t = template.Template(slot_content)
        content = t.render(context)
        return mark_safe(content)



from __future__ import annotations

from pathlib import Path

import pytest

from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType


class TestAsset:
    def test_hash(self):
        asset1 = Asset(Path("static.css"), AssetType.CSS)
        asset2 = Asset(Path("static.css"), AssetType.CSS)

        assert asset1 == asset2
        assert hash(asset1) == hash(asset2)

        assets = {asset1, asset2, Asset(Path("other.css"), AssetType.CSS)}

        assert len(assets) == 2

    def test_exists(self, tmp_path: Path):
        css_file = tmp_path / "test.css"
        css_file.touch()

        asset = Asset(css_file, AssetType.CSS)

        assert asset.exists() is True

    def test_exists_nonexistent(self):
        missing_asset = Asset(Path("missing.css"), AssetType.CSS)
        assert missing_asset.exists() is False

    @pytest.mark.parametrize(
        "asset,expected_html_tag_bits",
        [
            (Asset(Path("static.css"), AssetType.CSS), 'link rel="stylesheet" href='),
            (Asset(Path("static.js"), AssetType.JS), "script src="),
        ],
    )
    def test_render(self, asset, expected_html_tag_bits):
        rendered = asset.render()
        assert expected_html_tag_bits in rendered
        assert asset.url in rendered

    @pytest.mark.parametrize(
        "path,expected",
        [
            (Path("static.css"), Asset(Path("static.css"), AssetType.CSS)),
            (Path("static.js"), Asset(Path("static.js"), AssetType.JS)),
            (
                Path("nested/path/style.css"),
                Asset(Path("nested/path/style.css"), AssetType.CSS),
            ),
            (
                Path("./relative/script.js"),
                Asset(Path("./relative/script.js"), AssetType.JS),
            ),
            (Path("UPPERCASE.CSS"), Asset(Path("UPPERCASE.CSS"), AssetType.CSS)),
            (Path("mixed.Js"), Asset(Path("mixed.Js"), AssetType.JS)),
        ],
    )
    def test_from_path(self, path, expected):
        assert Asset.from_path(path) == expected

    def test_from_path_invalid(self):
        with pytest.raises(ValueError):
            Asset.from_path(Path("static.html"))



from __future__ import annotations

import pytest
from django.template.exceptions import TemplateSyntaxError

from django_bird.templatetags.tags.prop import parse_prop_name


@pytest.mark.parametrize(
    "bits,expected",
    [
        (["bird:prop", "id"], ("id", None)),
        (["bird:prop", "class='btn'", "foo"], ("class", "btn")),
    ],
)
def test_parse_prop_name(bits, expected):
    assert parse_prop_name(bits) == expected


def test_parse_prop_name_no_args():
    with pytest.raises(TemplateSyntaxError):
        parse_prop_name([])



from __future__ import annotations

import pytest
from django.test import override_settings

from django_bird.templates import get_template_names


@pytest.mark.parametrize(
    "name,component_dirs,expected",
    [
        (
            "button",
            [],
            [
                "bird/button/button.html",
                "bird/button/index.html",
                "bird/button.html",
            ],
        ),
        (
            "input.label",
            [],
            [
                "bird/input/label/label.html",
                "bird/input/label/index.html",
                "bird/input/label.html",
                "bird/input.label.html",
            ],
        ),
        (
            "button",
            ["custom", "theme"],
            [
                "custom/button/button.html",
                "custom/button/index.html",
                "custom/button.html",
                "theme/button/button.html",
                "theme/button/index.html",
                "theme/button.html",
                "bird/button/button.html",
                "bird/button/index.html",
                "bird/button.html",
            ],
        ),
    ],
)
def test_get_template_names(name, component_dirs, expected):
    with override_settings(DJANGO_BIRD={"COMPONENT_DIRS": component_dirs}):
        template_names = get_template_names(name)

    assert template_names == expected


def test_get_template_names_invalid():
    template_names = get_template_names("input.label")

    assert "bird/input/label/invalid.html" not in template_names


def test_get_template_names_duplicates():
    with override_settings(DJANGO_BIRD={"COMPONENT_DIRS": ["bird"]}):
        template_names = get_template_names("button")

        template_counts = {}
        for template in template_names:
            template_counts[template] = template_counts.get(template, 0) + 1

        for _, count in template_counts.items():
            assert count == 1



from __future__ import annotations

from django.urls import path

from .views import asset_view

app_name = "django_bird"

urlpatterns = [
    path("assets/<str:component_name>/<str:asset_filename>", asset_view, name="asset"),
]



from __future__ import annotations

import pytest
from django.template import Context
from django.template import Template
from django.template.exceptions import TemplateSyntaxError

from django_bird.templatetags.tags.slot import parse_slot_name
from tests.utils import TestComponent
from tests.utils import TestComponentCase


@pytest.mark.parametrize(
    "bits,expected",
    [
        (["slot"], "default"),
        (["slot", "foo"], "foo"),
        (["slot", "'foo'"], "foo"),
        (["slot", '"foo"'], "foo"),
        (["slot", 'name="foo"'], "foo"),
        (["slot", "name='foo'"], "foo"),
    ],
)
def test_parse_slot_name(bits, expected):
    assert parse_slot_name(bits) == expected


def test_parse_slot_name_no_args():
    with pytest.raises(TemplateSyntaxError):
        assert parse_slot_name([])


class TestTemplateTag:
    @pytest.mark.parametrize(
        "component_content",
        [
            "{{ slot }}",
            "{% bird:slot %}{% endbird:slot %}",
            "{% bird:slot default %}{% endbird:slot %}",
            "{% bird:slot 'default' %}{% endbird:slot %}",
            '{% bird:slot "default" %}{% endbird:slot %}',
            "{% bird:slot name=default %}{% endbird:slot %}",
            "{% bird:slot name='default' %}{% endbird:slot %}",
            '{% bird:slot name="default" %}{% endbird:slot %}',
        ],
    )
    def test_default_slot(self, component_content, templates_dir, normalize_whitespace):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content=component_content,
            ),
            template_content="{% bird test %}Content{% endbird %}",
            template_context={"slot": "Content"},
            expected="Content",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_default_content(self, templates_dir, normalize_whitespace):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content="""
                    <button>
                        {% bird:slot leading-icon %}
                            <span>Default icon</span>
                        {% endbird:slot %}

                        {% bird:slot %}
                            Click me
                        {% endbird:slot %}
                    </button>
                """,
            ),
            template_content="{% bird test %}{% endbird %}",
            expected="<button><span>Default icon</span>Click me</button>",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context({}))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_default_content_override(self, templates_dir, normalize_whitespace):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content="""
                    <button>
                        {% bird:slot leading-icon %}
                            <span>Default icon</span>
                        {% endbird:slot %}

                        {% bird:slot %}
                            Click me
                        {% endbird:slot %}
                    </button>
                """,
            ),
            template_content="""
                {% bird test %}
                    {% bird:slot leading-icon %}→{% endbird:slot %}
                    Submit
                {% endbird %}
            """,
            expected="<button>→ Submit</button>",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context({}))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "component_content",
        [
            "{{ slots.named_slot }}",
            "{% bird:slot named_slot %}{% endbird:slot %}",
            "{% bird:slot name=named_slot %}{% endbird:slot %}",
        ],
    )
    def test_named_slot(self, component_content, templates_dir, normalize_whitespace):
        test_case = TestComponentCase(
            component=TestComponent(
                name="test",
                content=component_content,
            ),
            template_content="""
                {% bird test %}
                    {% bird:slot named_slot %}{% endbird:slot %}
                {% endbird %}
            """,
            template_context={
                "slots": {
                    "named_slot": "Content",
                },
            },
            expected="Content",
        )
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Outer slot replacement",
                component=TestComponent(
                    name="test",
                    content="""
                        {% bird:slot outer %}
                            Outer {% bird:slot inner %}Inner{% endbird:slot %} Content
                        {% endbird:slot %}
                    """,
                ),
                template_content="""
                    {% bird test %}
                        {% bird:slot outer %}Replaced Content{% endbird:slot %}
                    {% endbird %}
                """,
                expected="Replaced Content",
            ),
            TestComponentCase(
                description="Inner slot replacement",
                component=TestComponent(
                    name="test",
                    content="""
                        {% bird:slot outer %}
                            Outer {% bird:slot inner %}Inner{% endbird:slot %} Content
                        {% endbird:slot %}
                    """,
                ),
                template_content="""
                    {% bird test %}
                        {% bird:slot inner %}Replaced Content{% endbird:slot %}
                    {% endbird %}
                """,
                expected="Outer Replaced Content Content",
            ),
            TestComponentCase(
                description="Both slots replaced",
                component=TestComponent(
                    name="test",
                    content="""
                        {% bird:slot outer %}
                            Outer {% bird:slot inner %}Inner{% endbird:slot %} Content
                        {% endbird:slot %}
                    """,
                ),
                template_content="""
                    {% bird test %}
                        {% bird:slot outer %}
                            New {% bird:slot inner %}Nested{% endbird:slot %} Text
                        {% endbird:slot %}
                    {% endbird %}
                """,
                expected="New Nested Text",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_nested_slots(self, test_case, templates_dir, normalize_whitespace):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    @pytest.mark.parametrize(
        "test_case",
        [
            TestComponentCase(
                description="Variable in slot",
                component=TestComponent(
                    name="test",
                    content="{% bird:slot %}{% endbird:slot %}",
                ),
                template_content="{% bird test %}{{ message }}{% endbird %}",
                template_context={"message": "Hello World"},
                expected="Hello World",
            ),
            TestComponentCase(
                description="Nested variable in slot",
                component=TestComponent(
                    name="test",
                    content="{% bird:slot %}{% endbird:slot %}",
                ),
                template_content="{% bird test %}{{ user.name }}{% endbird %}",
                template_context={"user": {"name": "John"}},
                expected="John",
            ),
            TestComponentCase(
                description="Template tag in slot",
                component=TestComponent(
                    name="test",
                    content="{% bird:slot %}{% endbird:slot %}",
                ),
                template_content="{% bird test %}{% if show %}Show{% endif %}{% endbird %}",
                template_context={"show": True},
                expected="Show",
            ),
        ],
        ids=lambda x: x.description,
    )
    def test_template_content(self, test_case, templates_dir, normalize_whitespace):
        test_case.component.create(templates_dir)

        template = Template(test_case.template_content)
        rendered = template.render(Context(test_case.template_context))

        assert normalize_whitespace(rendered) == test_case.expected

    def test_too_many_args(self):
        with pytest.raises(TemplateSyntaxError):
            Template("{% bird:slot too many args %}{% endbird:slot %}")



from __future__ import annotations

from contextlib import contextmanager

from django.template import Context
from django.template import Template

from django_bird.components import components
from django_bird.staticfiles import AssetType

from .utils import TestAsset
from .utils import TestComponent


@contextmanager
def discover_components():
    try:
        yield
    finally:
        components.discover_components()


def test_template_inheritance_assets(create_template, templates_dir):
    with discover_components():
        parent = TestComponent(name="parent", content="<div>Parent</div>").create(
            templates_dir
        )
        child = TestComponent(name="child", content="<div>Child</div>").create(
            templates_dir
        )

        parent_css = TestAsset(
            component=parent,
            content=".parent { color: red; }",
            asset_type=AssetType.CSS,
        ).create()
        child_css = TestAsset(
            component=child,
            content=".parent { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

    base_template = templates_dir / "base.html"
    base_template.write_text("""
        {% bird:css %}

        {% bird parent %}Parent Content{% endbird %}
        {% block content %}{% endblock %}
    """)

    template = Template("""
        {% extends 'base.html' %}

        {% block content %}
            {% bird child %}Child Content{% endbird %}
        {% endblock %}
    """)
    rendered = template.render(Context({}))

    assert str(parent_css.file.name) in rendered
    assert str(child_css.file.name) in rendered



from __future__ import annotations

import sys
from typing import TypeGuard

from django.template.base import Node
from django.template.base import Template

if sys.version_info >= (3, 12):
    from typing import override as typing_override
else:
    from typing_extensions import (
        override as typing_override,  # pyright: ignore[reportUnreachable]
    )

override = typing_override

TagBits = list[str]


def _has_nodelist(node: Template | Node) -> TypeGuard[Template]:
    return hasattr(node, "nodelist")



# pyright: reportAny=false
from __future__ import annotations

from typing import Any

from django import template
from django.template.base import NodeList
from django.template.base import Parser
from django.template.base import Token
from django.template.context import Context

from django_bird._typing import TagBits
from django_bird._typing import override
from django_bird.components import Component
from django_bird.components import components
from django_bird.conf import app_settings
from django_bird.params import Param
from django_bird.params import Params
from django_bird.params import Value
from django_bird.slots import DEFAULT_SLOT
from django_bird.slots import Slots

TAG = "bird"
END_TAG = "endbird"


def do_bird(parser: Parser, token: Token) -> BirdNode:
    bits = token.split_contents()
    if len(bits) == 1:
        msg = f"{TAG} tag requires at least one argument"
        raise template.TemplateSyntaxError(msg)

    name = bits[1]
    attrs = []
    isolated_context = False

    for bit in bits[2:]:
        match bit:
            case "only":
                isolated_context = True
            case "/":
                continue
            case _:
                param = Param.from_bit(bit)
                attrs.append(param)

    nodelist = parse_nodelist(bits, parser)
    return BirdNode(name, attrs, nodelist, isolated_context)


def parse_nodelist(bits: TagBits, parser: Parser) -> NodeList | None:
    # self-closing tag
    # {% bird name / %}
    if len(bits) > 0 and bits[-1] == "/":
        nodelist = None
    else:
        nodelist = parser.parse((END_TAG,))
        parser.delete_first_token()
    return nodelist


class BirdNode(template.Node):
    def __init__(
        self,
        name: str,
        attrs: list[Param],
        nodelist: NodeList | None,
        isolated_context: bool = False,
    ) -> None:
        self.name = name
        self.attrs = attrs
        self.nodelist = nodelist
        self.isolated_context = isolated_context

    @override
    def render(self, context: Context) -> str:
        component_name = self.get_component_name(context)
        component = components.get_component(component_name)
        component_context = self.get_component_context_data(component, context)

        if self.isolated_context:
            return component.render(context.new(component_context))
        with context.push(**component_context):
            return component.render(context)

    def get_component_name(self, context: Context) -> str:
        try:
            name = template.Variable(self.name).resolve(context)
        except template.VariableDoesNotExist:
            name = self.name
        return name

    def get_component_context_data(
        self, component: Component, context: Context
    ) -> dict[str, Any]:
        if app_settings.ENABLE_BIRD_ID_ATTR:
            self.attrs.append(Param("data_bird_id", Value(component.id, True)))

        params = Params.with_attrs(self.attrs)
        props = params.render_props(component.nodelist, context)
        attrs = params.render_attrs(context)
        slots = Slots.collect(self.nodelist, context).render()
        default_slot = slots.get(DEFAULT_SLOT) or context.get("slot")

        return {
            "attrs": attrs,
            "props": props,
            "slot": default_slot,
            "slots": slots,
        }



from __future__ import annotations

import pytest
from django.template.base import Node
from django.template.base import NodeList
from django.template.base import Template
from django.template.context import Context
from django.template.engine import Engine
from django.template.loader import get_template

from django_bird.components import components
from django_bird.loader import BIRD_TAG_PATTERN
from django_bird.loader import BirdLoader
from django_bird.staticfiles import AssetType
from django_bird.templatetags.tags.bird import BirdNode

from .utils import TestAsset
from .utils import TestComponent


@pytest.mark.parametrize(
    "template_content,expected_matches",
    [
        ("{% bird button %}", ["button"]),
        ("{% bird alert %}{% bird button %}", ["alert", "button"]),
        ("{% bird 'button' %}", ["button"]),
        ('{% bird "button" %}', ["button"]),
        ("{%bird button%}", ["button"]),
        ("{%    bird    button    %}", ["button"]),
        ("{% endbird button %}", []),
        ("{% birds button %}", []),
        ("<bird:button>", []),
        ("{% extends 'base.html' %}{% bird button %}", ["button"]),
    ],
)
def test_bird_tag_pattern(template_content, expected_matches):
    matches = [
        m.group(1).strip("'\"") for m in BIRD_TAG_PATTERN.finditer(template_content)
    ]
    assert matches == expected_matches


@pytest.mark.parametrize(
    "template_name",
    (
        "with_bird.html",
        "without_bird.html",
    ),
)
def test_render_template(template_name):
    context = {
        "title": "Test Title",
        "content": "Test Content",
    }

    template = get_template(template_name)
    rendered = template.render(context)

    assert rendered


@pytest.mark.parametrize(
    "node,expected_count",
    [
        (Template("{% bird button %}Click me{% endbird %}"), 1),
        (BirdNode(name="button", attrs=[], nodelist=None), 1),
        (
            BirdNode(
                name="button",
                attrs=[],
                nodelist=NodeList(
                    [BirdNode(name="button", attrs=[], nodelist=None)],
                ),
            ),
            1,
        ),
        (type("NodeWithNoneNodelist", (Node,), {"nodelist": None})(), 0),
        (Template("{% bird button %}{% bird button %}{% endbird %}{% endbird %}"), 1),
    ],
)
def test_ensure_components_loaded(node, expected_count, templates_dir):
    button = TestComponent(name="button", content="<button>Click me</button>").create(
        templates_dir
    )

    TestAsset(
        component=button, content=".button { color: blue; }", asset_type=AssetType.CSS
    ).create()
    TestAsset(
        component=button, content="console.log('button');", asset_type=AssetType.JS
    ).create()

    loader = BirdLoader(Engine.get_default())
    context = Context()

    loader._ensure_components_loaded(node, context)

    assert len(components._components) == expected_count



from __future__ import annotations

from dataclasses import dataclass

from django.template.base import NodeList
from django.template.base import Template
from django.template.context import Context

DEFAULT_SLOT = "default"


@dataclass
class Slots:
    slots: dict[str, list[str]]
    context: Context

    @classmethod
    def collect(cls, nodelist: NodeList | None, context: Context):
        from django_bird.templatetags.tags.slot import SlotNode

        if nodelist is None:
            return cls({}, context)

        slots: dict[str, list[str]] = {DEFAULT_SLOT: []}
        active_slot = DEFAULT_SLOT

        for node in nodelist:
            if isinstance(node, SlotNode):
                active_slot = node.name
                slots.setdefault(active_slot, [])
            else:
                active_slot = DEFAULT_SLOT

            rendered_content = node.render(context)
            slots[active_slot].append(rendered_content)

        if all(not content for content in slots[DEFAULT_SLOT]) and "slot" in context:
            slots[DEFAULT_SLOT] = [context["slot"]]

        return cls(slots, context)

    def render(self):
        return {
            slot: Template("".join(content)).render(self.context)
            for slot, content in self.slots.items()
        }



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any

from django import template
from django.template.base import NodeList
from django.template.context import Context
from django.utils.safestring import SafeString
from django.utils.safestring import mark_safe

from .templatetags.tags.prop import PropNode


@dataclass
class Value:
    raw: str | bool | None
    quoted: bool = False

    def resolve(self, context: Context | dict[str, Any]) -> Any:
        if self.raw is None or (isinstance(self.raw, str) and self.raw == "False"):
            return None
        if (isinstance(self.raw, bool) and self.raw) or (
            isinstance(self.raw, str) and self.raw == "True"
        ):
            return True
        if isinstance(self.raw, str) and not self.quoted:
            try:
                return template.Variable(str(self.raw)).resolve(context)
            except template.VariableDoesNotExist:
                return self.raw
        return self.raw


@dataclass
class Param:
    name: str
    value: Value

    def render_attr(self, context: Context) -> str:
        value = self.value.resolve(context)
        name = self.name.replace("_", "-")
        if value is None:
            return ""
        if value is True:
            return name
        return f'{name}="{value}"'

    def render_prop(self, context: Context) -> str | bool | None:
        return self.value.resolve(context)

    @classmethod
    def from_bit(cls, bit: str) -> Param:
        if "=" in bit:
            name, raw_value = bit.split("=", 1)
            # Check if the value is quoted
            if raw_value.startswith(("'", '"')) and raw_value.endswith(raw_value[0]):
                value = Value(raw_value[1:-1], quoted=True)
            else:
                value = Value(raw_value.strip(), quoted=False)
        else:
            name, value = bit, Value(True)
        return cls(name, value)


@dataclass
class Params:
    attrs: list[Param] = field(default_factory=list)
    props: list[Param] = field(default_factory=list)

    @classmethod
    def with_attrs(cls, attrs: list[Param] | None) -> Params:
        """Create a Params instance with a copy of the provided attrs."""
        return cls(attrs=attrs.copy() if attrs is not None else [], props=[])

    def render_props(self, nodelist: NodeList | None, context: Context):
        if nodelist is None:
            return

        attrs_to_remove = set()

        for node in nodelist:
            if not isinstance(node, PropNode):
                continue

            value = Value(node.default, quoted=False)

            for idx, attr in enumerate(self.attrs):
                if node.name == attr.name:
                    resolved = attr.value.resolve(context)
                    if resolved is not None:
                        value = attr.value
                    attrs_to_remove.add(idx)

            self.props.append(Param(name=node.name, value=value))

        for idx in sorted(attrs_to_remove, reverse=True):
            self.attrs.pop(idx)

        return {prop.name: prop.render_prop(context) for prop in self.props}

    def render_attrs(self, context: Context) -> SafeString:
        rendered = " ".join(attr.render_attr(context) for attr in self.attrs)
        return mark_safe(rendered)



from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Any

import django.template
from django.conf import settings

from ._typing import override

DJANGO_BIRD_SETTINGS_NAME = "DJANGO_BIRD"


@dataclass
class AppSettings:
    COMPONENT_DIRS: list[Path | str] = field(default_factory=list)
    ENABLE_AUTO_CONFIG: bool = True
    ENABLE_BIRD_ID_ATTR: bool = True
    _template_configurator: TemplateConfigurator = field(init=False)

    def __post_init__(self):
        self._template_configurator = TemplateConfigurator(self)

    @override
    def __getattribute__(self, __name: str) -> object:
        user_settings = getattr(settings, DJANGO_BIRD_SETTINGS_NAME, {})
        return user_settings.get(__name, super().__getattribute__(__name))  # pyright: ignore[reportAny]

    def autoconfigure(self) -> None:
        if not self.ENABLE_AUTO_CONFIG:
            return

        self._template_configurator.autoconfigure()


class TemplateConfigurator:
    def __init__(self, app_settings: AppSettings, engine_name: str = "django") -> None:
        self.app_settings = app_settings
        self.engine_name = engine_name
        self._configured = False

    def autoconfigure(self) -> None:
        for template_config in settings.TEMPLATES:
            engine_name = (
                template_config.get("NAME") or template_config["BACKEND"].split(".")[-2]
            )
            if engine_name != self.engine_name:
                return

        options = template_config.setdefault("OPTIONS", {})

        self.configure_loaders(options)
        self.configure_builtins(options)

        # Force re-evaluation of settings.TEMPLATES because EngineHandler caches it.
        with suppress(AttributeError):
            del django.template.engines.templates
            django.template.engines._engines = {}  # type: ignore[attr-defined]

        self._configured = True

    def configure_loaders(self, options: dict[str, Any]) -> None:
        loaders = options.setdefault("loaders", [])

        # find the inner-most loaders, which is an iterable of only strings
        while not all(isinstance(loader, str) for loader in loaders):
            for loader in loaders:
                # if we've found a list or tuple, we aren't yet in the inner-most loaders
                if isinstance(loader, list | tuple):
                    # reassign `loaders` variable to force the while loop restart
                    loaders = loader

        # if django-bird's loader is the first, we good
        loaders_already_configured = (
            len(loaders) > 0 and "django_bird.loader.BirdLoader" == loaders[0]
        )

        if not loaders_already_configured:
            loaders.insert(0, "django_bird.loader.BirdLoader")

    def configure_builtins(self, options: dict[str, Any]) -> None:
        builtins = options.setdefault("builtins", [])

        builtins_already_configured = "django_bird.templatetags.django_bird" in builtins

        if not builtins_already_configured:
            builtins.append("django_bird.templatetags.django_bird")


app_settings = AppSettings()



from __future__ import annotations

import pytest
from django.template.base import Parser
from django.template.base import Token
from django.template.base import TokenType
from django.template.context import Context
from django.template.exceptions import TemplateSyntaxError

from django_bird.staticfiles import AssetType
from django_bird.templatetags.tags.asset import CSS_TAG
from django_bird.templatetags.tags.asset import JS_TAG
from django_bird.templatetags.tags.asset import AssetNode
from django_bird.templatetags.tags.asset import do_asset
from tests.utils import TestAsset
from tests.utils import TestComponent


class TestTemplateTag:
    @pytest.mark.parametrize(
        "tag,expected",
        [
            (CSS_TAG, AssetType.CSS),
            (JS_TAG, AssetType.JS),
        ],
    )
    def test_asset_type(self, tag, expected):
        token = Token(TokenType.BLOCK, tag)
        parser = Parser([])
        node = do_asset(parser, token)
        assert node.asset_type == expected

    def test_missing_tag_name(self):
        token = Token(TokenType.BLOCK, "")
        parser = Parser([])
        with pytest.raises(TemplateSyntaxError):
            do_asset(parser, token)

    @pytest.mark.parametrize(
        "tag",
        ["bird:jsx", "birdcss"],
    )
    def test_invalid_tag_name(self, tag):
        token = Token(TokenType.BLOCK, tag)
        parser = Parser([])
        with pytest.raises(ValueError):
            do_asset(parser, token)

    def test_template_inheritence(self, create_template, templates_dir):
        alert = TestComponent(
            name="alert", content='<div class="alert">{{ slot }}</div>'
        ).create(templates_dir)
        alert_css = TestAsset(
            component=alert, content=".alert { color: red; }", asset_type=AssetType.CSS
        ).create()
        alert_js = TestAsset(
            component=alert, content="console.log('alert');", asset_type=AssetType.JS
        ).create()

        button = TestComponent(
            name="button", content="<button>{{ slot }}</button>"
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        button_js = TestAsset(
            component=button, content="console.log('button');", asset_type=AssetType.JS
        ).create()

        base_path = templates_dir / "base.html"
        base_path.write_text("""
            <html>
            <head>
                <title>Test</title>
                {% bird:css %}
            </head>
            <body>
                {% bird alert %}Base Alert{% endbird %}
                {% block content %}{% endblock %}
                {% bird:js %}
            </body>
            </html>
        """)

        child_path = templates_dir / "child.html"
        child_path.write_text("""
            {% extends 'base.html' %}
            {% block content %}
                {% bird button %}Click me{% endbird %}
            {% endblock %}
        """)

        template = create_template(child_path)

        rendered = template.render({})

        assert (
            f'<link rel="stylesheet" href="/__bird__/assets/{alert.name}/{alert_css.file.name}">'
            in rendered
        )
        assert (
            f'<link rel="stylesheet" href="/__bird__/assets/{button.name}/{button_css.file.name}">'
            in rendered
        )
        assert (
            f'<script src="/__bird__/assets/{alert.name}/{alert_js.file.name}"></script>'
            in rendered
        )
        assert (
            f'<script src="/__bird__/assets/{button.name}/{button_js.file.name}"></script>'
            in rendered
        )

    def test_with_no_assets(self, create_template, templates_dir):
        TestComponent(
            name="alert", content='<div class="alert">{{ slot }}</div>'
        ).create(templates_dir)

        base_path = templates_dir / "base.html"
        base_path.write_text("""
            <html>
            <head>
                <title>Test</title>
                {% bird:css %}
            </head>
            <body>
                {% bird alert %}Base Alert{% endbird %}
                {% block content %}{% endblock %}
                {% bird:js %}
            </body>
            </html>
        """)

        template = create_template(base_path)

        rendered = template.render({})

        assert '<link rel="stylesheet" href="' not in rendered
        assert '<script src="' not in rendered

    def test_component_render_order(self, create_template, templates_dir):
        first = TestComponent(
            name="first", content="<div>First: {{ slot }}</div>"
        ).create(templates_dir)
        first_css = TestAsset(
            component=first, content=".first { color: red; }", asset_type=AssetType.CSS
        ).create()
        first_js = TestAsset(
            component=first, content="console.log('first');", asset_type=AssetType.JS
        ).create()

        second = TestComponent(
            name="second", content="<div>Second: {{ slot }}</div>"
        ).create(templates_dir)
        second_css = TestAsset(
            component=second,
            content=".second { color: red; }",
            asset_type=AssetType.CSS,
        ).create()
        second_js = TestAsset(
            component=second, content="console.log('second');", asset_type=AssetType.JS
        ).create()

        template_path = templates_dir / "test.html"
        template_path.write_text("""
            <html>
            <head>
                {% bird:css %}
            </head>
            <body>
                <div>
                    {% bird first %}One{% endbird %}
                    <p>Some content</p>
                    {% bird second %}Two{% endbird %}
                </div>
                {% bird:js %}
            </body>
            </html>
        """)

        template = create_template(template_path)

        rendered = template.render({})

        head_end = rendered.find("</head>")
        assert (
            f'<link rel="stylesheet" href="/__bird__/assets/{first.name}/{first_css.file.name}">'
            in rendered[:head_end]
        )
        assert (
            f'<link rel="stylesheet" href="/__bird__/assets/{second.name}/{second_css.file.name}">'
            in rendered[:head_end]
        )

        body_start = rendered.find("<body")
        assert (
            f'<script src="/__bird__/assets/{first.name}/{first_js.file.name}"></script>'
            in rendered[body_start:]
        )
        assert (
            f'<script src="/__bird__/assets/{second.name}/{second_js.file.name}"></script>'
            in rendered[body_start:]
        )

    def test_asset_duplication(self, create_template, templates_dir):
        alert = TestComponent(
            name="alert", content='<div class="alert">{{ slot }}</div>'
        ).create(templates_dir)
        alert_css = TestAsset(
            component=alert, content=".alert { color: red; }", asset_type=AssetType.CSS
        ).create()
        alert_js = TestAsset(
            component=alert, content="console.log('alert');", asset_type=AssetType.JS
        ).create()

        base_path = templates_dir / "base.html"
        base_path.write_text("""
            <html>
            <head>
                <title>Test</title>
                {% bird:css %}
            </head>
            <body>
                {% bird alert %}Base Alert{% endbird %}
                {% bird alert %}Base Alert{% endbird %}
                {% bird alert %}Base Alert{% endbird %}
                {% block content %}{% endblock %}
                {% bird:js %}
            </body>
            </html>
        """)

        child_path = templates_dir / "child.html"
        child_path.write_text("""
            {% extends 'base.html' %}
            {% block content %}
                {% bird alert %}Base Alert{% endbird %}
                {% bird alert %}Base Alert{% endbird %}
                {% bird alert %}Base Alert{% endbird %}
            {% endblock %}
        """)

        template = create_template(child_path)

        rendered = template.render({})

        assert (
            rendered.count(
                f'<link rel="stylesheet" href="/__bird__/assets/{alert.name}/{alert_css.file.name}">'
            )
            == 1
        )
        assert (
            rendered.count(
                f'<script src="/__bird__/assets/{alert.name}/{alert_js.file.name}"></script>'
            )
            == 1
        )

    def test_unused_component_asset_not_rendered(self, create_template, templates_dir):
        alert = TestComponent(
            name="alert", content='<div class="alert">{{ slot }}</div>'
        ).create(templates_dir)
        alert_css = TestAsset(
            component=alert, content=".alert { color: red; }", asset_type=AssetType.CSS
        ).create()
        alert_js = TestAsset(
            component=alert, content="console.log('alert');", asset_type=AssetType.JS
        ).create()

        button = TestComponent(
            name="button", content="<button>{{ slot }}</button>"
        ).create(templates_dir)
        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        button_js = TestAsset(
            component=button, content="console.log('button');", asset_type=AssetType.JS
        ).create()

        base_path = templates_dir / "base.html"
        base_path.write_text("""
            <html>
            <head>
                <title>Test</title>
                {% bird:css %}
            </head>
            <body>
                {% bird alert %}Base Alert{% endbird %}
                {% block content %}{% endblock %}
                {% bird:js %}
            </body>
            </html>
        """)

        child_path = templates_dir / "child.html"
        child_path.write_text("""
            {% extends 'base.html' %}
            {% block content %}
                {% bird alert %}Base Alert{% endbird %}
            {% endblock %}
        """)

        template = create_template(child_path)

        rendered = template.render({})

        assert (
            f'<link rel="stylesheet" href="/__bird__/assets/{alert.name}/{alert_css.file.name}">'
            in rendered
        )
        assert (
            f'<script src="/__bird__/assets/{alert.name}/{alert_js.file.name}"></script>'
            in rendered
        )
        assert (
            f'<link rel="stylesheet" href="/__bird__/assets/{button.name}/{button_css.file.name}">'
            not in rendered
        )
        assert (
            f'<script src="/__bird__/assets/{button.name}/{button_js.file.name}"></script>'
            not in rendered
        )


class TestNode:
    def test_no_template(self):
        node = AssetNode(AssetType.CSS)
        context = Context({})
        assert node.render(context) == ""



from __future__ import annotations

import queue
import shutil
import sys
import tempfile
import threading
import time
from pathlib import Path

import pytest
from django.template.backends.django import Template
from django.template.context import Context
from django.template.exceptions import TemplateDoesNotExist
from django.test import override_settings

from django_bird.components import Component
from django_bird.components import ComponentRegistry
from django_bird.components import components
from django_bird.staticfiles import Asset
from django_bird.staticfiles import AssetType

from .utils import TestAsset
from .utils import TestComponent


class TestComponentClass:
    def test_from_name_basic(self, templates_dir):
        TestComponent(name="button", content="<button>Click me</button>").create(
            templates_dir
        )

        comp = Component.from_name("button")

        assert comp.name == "button"
        assert isinstance(comp.template, Template)
        assert comp.render(Context({})) == "<button>Click me</button>"

    def test_from_name_with_assets(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()
        button_js = TestAsset(
            component=button, content="console.log('button');", asset_type=AssetType.JS
        ).create()

        comp = Component.from_name("button")

        assert comp.name == "button"
        assert len(comp.assets) == 2
        assert Asset(button_css.file, AssetType.CSS) in comp.assets
        assert Asset(button_js.file, AssetType.JS) in comp.assets

    def test_from_name_with_partial_assets_css(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            button, ".button { color: blue; }", asset_type=AssetType.CSS
        ).create()

        comp = Component.from_name("button")

        assert comp.name == "button"
        assert len(comp.assets) == 1
        assert Asset(button_css.file, button_css.asset_type) in comp.assets

    def test_from_name_with_partial_assets_js(self, templates_dir):
        button = TestComponent(name="button", content="<button>Click me</button>")
        button.create(templates_dir)

        button_js = TestAsset(
            button, "console.log('button');", asset_type=AssetType.JS
        ).create()

        comp = Component.from_name("button")

        assert comp.name == "button"
        assert len(comp.assets) == 1
        assert Asset(button_js.file, button_js.asset_type) in comp.assets

    def test_from_name_no_assets(self, templates_dir):
        button = TestComponent(name="button", content="<button>Click me</button>")
        button.create(templates_dir)

        comp = Component.from_name("button")

        assert len(comp.assets) == 0

    def test_from_name_custom_component_dir(self, templates_dir, override_app_settings):
        TestComponent(
            name="button", content="<button>Click me</button>", parent_dir="components"
        ).create(templates_dir)

        with override_app_settings(COMPONENT_DIRS=["components"]):
            comp = Component.from_name("button")

        assert comp.name == "button"
        assert isinstance(comp.template, Template)
        assert comp.render(Context({})) == "<button>Click me</button>"

    def test_id_is_consistent(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        comp1 = Component.from_name(button.name)
        comp2 = Component.from_name(button.name)

        assert comp1.id == comp2.id

    def test_id_content_changes(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        comp1 = Component.from_name(button.name)

        button.file.write_text("<button>Don't click me</button>")
        comp2 = Component.from_name(button.name)

        assert comp1.id != comp2.id

    def test_id_whitespace_changes(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        comp1 = Component.from_name(button.name)

        button.file.write_text("<button>\n  Click me\n</button>")
        comp2 = Component.from_name(button.name)

        assert comp1.id == comp2.id

    def test_id_name_changes(self, templates_dir):
        button1 = TestComponent(
            name="button1", content="<button>Click me</button>"
        ).create(templates_dir)
        button2 = TestComponent(
            name="button2", content="<button>Click me</button>"
        ).create(templates_dir)

        comp1 = Component.from_name(button1.name)
        comp2 = Component.from_name(button2.name)

        assert comp1.id != comp2.id

    def test_id_formatting(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        comp = Component.from_name(button.name)

        assert len(comp.id) == 7
        assert all(c in "0123456789abcdef" for c in comp.id)


class TestComponentRegistryProject:
    def test_discover_components(self, templates_dir):
        TestComponent(name="button", content="<button>Click me</button>").create(
            templates_dir
        )
        TestComponent(name="alert", content="<div>Alert</div>").create(templates_dir)

        components.discover_components()

        assert "button" in components._components
        assert "alert" in components._components

    def test_custom_dir(self, templates_dir, override_app_settings):
        TestComponent(
            name="button", content="<button>Click me</button>", parent_dir="components"
        ).create(templates_dir)

        with override_app_settings(COMPONENT_DIRS=["components"]):
            components.discover_components()

        assert "button" in components._components

    def test_missing_dir(self, override_app_settings):
        with override_app_settings(COMPONENT_DIRS=["nonexistent"]):
            components.discover_components()

        assert len(components._components) == 0

    def test_non_html_files(self, templates_dir):
        (templates_dir / "README.md").write_text("# Components")
        (templates_dir / "ignored.txt").write_text("Not a component")

        components.discover_components()

        assert len(components._components) == 0

    def test_empty_dirs(self, templates_dir):
        bird_dir = templates_dir / "bird"
        bird_dir.mkdir(parents=True)
        (bird_dir / "empty").mkdir()

        components.discover_components()

        assert len(components._components) == 0

    def test_nested_components(self, templates_dir):
        TestComponent(
            name="button",
            content="<button>Nested</button>",
            sub_dir="nested",
        ).create(templates_dir)

        components.discover_components()

        assert "nested.button" in components._components

    def test_component_name_collision(self, templates_dir, override_app_settings):
        TestComponent(
            name="button", content="<button>First</button>", parent_dir="first"
        ).create(templates_dir)

        TestComponent(
            name="button", content="<button>Second</button>", parent_dir="second"
        ).create(templates_dir)

        with override_app_settings(COMPONENT_DIRS=["first", "second"]):
            components.discover_components()
            component = components.get_component("button")
            assert "First" in component.template.template.source
            assert "Second" not in component.template.template.source

        components.clear()

        with override_app_settings(COMPONENT_DIRS=["second", "first"]):
            components.discover_components()
            component = components.get_component("button")
            assert "Second" in component.template.template.source
            assert "First" not in component.template.template.source


class TestComponentRegistryApps:
    @classmethod
    def setup_class(cls):
        cls.tmp_dir = tempfile.mkdtemp()
        cls.tmp_path = Path(cls.tmp_dir)

        sys.path.insert(0, str(cls.tmp_path))

        cls.app1_dir = cls.tmp_path / "test_app1"
        cls.app2_dir = cls.tmp_path / "test_app2"

        for app_dir in [cls.app1_dir, cls.app2_dir]:
            templates_dir = app_dir / "templates"
            templates_dir.mkdir(parents=True)

            (app_dir / "__init__.py").touch()

            (app_dir / "apps.py").write_text(f"""
from django.apps import AppConfig

class {app_dir.name.capitalize()}Config(AppConfig):
    name = '{app_dir.name}'
    path = '{str(app_dir)}'
""")

    @classmethod
    def teardown_class(cls):
        sys.path.remove(str(cls.tmp_path))
        shutil.rmtree(cls.tmp_dir)

    def test_discover_components_from_app(self):
        app1_templates = self.app1_dir / "templates"

        TestComponent(name="app1_button", content="<button>App1</button>").create(
            app1_templates
        )

        with override_settings(INSTALLED_APPS=["test_app1"]):
            components.discover_components()

        assert "app1_button" in components._components

    def test_multiple_apps_components(self):
        TestComponent(name="app1_button", content="<button>App1</button>").create(
            self.app1_dir / "templates"
        )

        TestComponent(name="app2_button", content="<button>App2</button>").create(
            self.app2_dir / "templates"
        )

        with override_settings(INSTALLED_APPS=["test_app1", "test_app2"]):
            components.discover_components()

        assert "app1_button" in components._components
        assert "app2_button" in components._components

    def test_app_component_override_project(self, templates_dir):
        TestComponent(name="button", content="<button>Project</button>").create(
            templates_dir
        )
        TestComponent(name="button", content="<button>App</button>").create(
            self.app1_dir / "templates"
        )

        with override_settings(INSTALLED_APPS=["test_app1"]):
            components.discover_components()
            component = components.get_component("button")
            assert "Project" in component.template.template.source
            assert "App" not in component.template.template.source

    def test_app_order_precedence(self):
        TestComponent(name="button", content="<button>App1</button>").create(
            self.app1_dir / "templates"
        )
        TestComponent(name="button", content="<button>App2</button>").create(
            self.app2_dir / "templates"
        )

        with override_settings(INSTALLED_APPS=["test_app1", "test_app2"]):
            components.discover_components()
            component = components.get_component("button")
            assert "App1" in component.template.template.source
            assert "App2" not in component.template.template.source

        components.clear()

        with override_settings(INSTALLED_APPS=["test_app2", "test_app1"]):
            components.discover_components()
            component = components.get_component("button")
            assert "App2" in component.template.template.source
            assert "App1" not in component.template.template.source


class TestComponentRegistryAssets:
    def test_discover_component_with_css(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: red; }",
            asset_type=AssetType.CSS,
        ).create()

        components.discover_components()

        assert "button" in components._components

        component = components._components["button"]

        assert len(component.assets) == 1
        assert Asset(button_css.file, button_css.asset_type) in component.assets

    def test_discover_component_with_js(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_js = TestAsset(
            component=button,
            content="console.log('clicked');",
            asset_type=AssetType.JS,
        ).create()

        components.discover_components()

        assert "button" in components._components

        component = components._components["button"]

        assert len(component.assets) == 1
        assert Asset(button_js.file, button_js.asset_type) in component.assets

    def test_get_assets_by_type(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: red; }",
            asset_type=AssetType.CSS,
        ).create()

        button_js = TestAsset(
            component=button,
            content="console.log('loaded');",
            asset_type=AssetType.JS,
        ).create()

        components.discover_components()

        css_assets = components.get_assets(AssetType.CSS)
        js_assets = components.get_assets(AssetType.JS)

        assert len(css_assets) == 1
        assert len(js_assets) == 1
        assert Asset(button_css.file, button_css.asset_type) in css_assets
        assert Asset(button_js.file, button_js.asset_type) in js_assets

    def test_assets_from_multiple_components(self, templates_dir):
        button1 = TestComponent(name="button1", content="<button>One</button>").create(
            templates_dir
        )
        button2 = TestComponent(name="button2", content="<button>Two</button>").create(
            templates_dir
        )

        button1_css = TestAsset(
            component=button1,
            content=".button1 { color: red; }",
            asset_type=AssetType.CSS,
        ).create()
        button2_css = TestAsset(
            component=button2,
            content=".button2 { color: blue; }",
            asset_type=AssetType.CSS,
        ).create()

        components.discover_components()

        css_assets = components.get_assets(AssetType.CSS)

        assert len(css_assets) == 2
        assert Asset(button1_css.file, button1_css.asset_type) in css_assets
        assert Asset(button2_css.file, button2_css.asset_type) in css_assets

    def test_missing_asset_file(self, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        TestAsset(
            component=button,
            content="/* Missing */",
            asset_type=AssetType.CSS,
        )

        components.discover_components()

        assert "button" in components._components

        component = components._components["button"]

        assert len(component.assets) == 0


class TestComponentRegistryCaching:
    def test_debug_mode_caching(self, templates_dir):
        component = TestComponent(
            name="button", content="<button>Original</button>"
        ).create(templates_dir)

        with override_settings(DEBUG=True):
            first = components.get_component("button")

            assert "button" in components._components
            assert "Original" in first.template.template.source

            component.file.write_text("<button>Updated</button>")
            second = components.get_component("button")

            assert second is not first
            assert "Updated" in second.template.template.source

    def test_production_mode_caching(self, templates_dir):
        component = TestComponent(
            name="button", content="<button>Original</button>"
        ).create(templates_dir)

        with override_settings(DEBUG=False):
            first = components.get_component("button")

            component.file.write_text("<button>Updated</button>")
            second = components.get_component("button")

            assert second is first
            assert "Original" in second.template.template.source

    @pytest.mark.parametrize("debug", [True, False])
    def test_asset_tracking(self, debug, templates_dir):
        button = TestComponent(
            name="button", content="<button>Click me</button>"
        ).create(templates_dir)

        button_css = TestAsset(
            component=button,
            content=".button { color: red; }",
            asset_type=AssetType.CSS,
        ).create()

        with override_settings(DEBUG=debug):
            components.get_component("button")
            css_assets = components.get_assets(AssetType.CSS)

            assert len(css_assets) == 1
            assert Asset(button_css.file, button_css.asset_type) in css_assets

    def test_lru_cache_limit(self, templates_dir):
        small_registry = ComponentRegistry(maxsize=2)

        for i in range(3):
            TestComponent(
                name=f"button{i}", content=f"<button>Button {i}</button>"
            ).create(templates_dir)

        small_registry.get_component("button0")
        small_registry.get_component("button1")
        small_registry.get_component("button2")  # This should evict button0

        assert "button0" not in small_registry._components
        assert "button1" in small_registry._components
        assert "button2" in small_registry._components

    def test_cache_clear(self, templates_dir):
        TestComponent(name="button", content="<button>Clear Me</button>").create(
            templates_dir
        )

        with override_settings(DEBUG=False):
            components.get_component("button")
            assert "button" in components._components

            components.clear()
            assert len(components._components) == 0

    def test_lru_cache_eviction_order(self, templates_dir):
        small_registry = ComponentRegistry(maxsize=2)

        for i in range(3):
            TestComponent(
                name=f"button{i}", content=f"<button>Button {i}</button>"
            ).create(templates_dir)

        small_registry.get_component("button0")
        small_registry.get_component("button1")

        # Access button0 again to make it more recently used than button1
        small_registry.get_component("button0")

        # Access button2, should evict button1 (least recently used)
        small_registry.get_component("button2")

        assert "button0" in small_registry._components
        assert "button1" not in small_registry._components
        assert "button2" in small_registry._components

    def test_concurrent_cache_access(self, templates_dir):
        TestComponent(name="shared_button", content="<button>Shared</button>").create(
            templates_dir
        )

        results = queue.Queue()

        def access_component():
            try:
                component = components.get_component("shared_button")
                results.put(("success", component.template.template.source))
            except Exception as e:
                results.put(("error", str(e)))

        # Create multiple threads to access the cache concurrently
        threads = [threading.Thread(target=access_component) for _ in range(5)]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        # Check that all accesses were successful
        while not results.empty():
            status, result = results.get()
            assert status == "success"
            assert "Shared" in result

    def test_cache_memory_limit(self, templates_dir):
        large_registry = ComponentRegistry(maxsize=5)
        large_content = "x" * 1_000_000  # 1MB of content

        for i in range(6):
            TestComponent(
                name=f"large_button{i}", content=f"<button>{large_content}</button>"
            ).create(templates_dir)

            large_registry.get_component(f"large_button{i}")

        assert len(large_registry._components) <= 5


class TestComponentRegistryErrors:
    def test_invalid_template_syntax(self, templates_dir):
        invalid_template = templates_dir / "invalid.html"
        invalid_template.write_text("{% invalid syntax %}")

        components.discover_components()

        with pytest.raises(TemplateDoesNotExist):
            components.get_component("invalid")

    def test_missing_required_template(self):
        components.discover_components()

        with pytest.raises(TemplateDoesNotExist):
            components.get_component("nonexistent")

    def test_invalid_asset_reference(self, templates_dir):
        button = TestComponent(name="button", content="<button>Error</button>").create(
            templates_dir
        )

        TestAsset(
            component=button,
            content="/* Invalid */",
            asset_type="invalid_type",
        )

        components.discover_components()

        assert "button" in components._components

        component = components._components["button"]

        assert len(component.assets) == 0


class TestComponentRegistryPerformance:
    def test_large_directory_scan(self, templates_dir):
        for i in range(100):
            TestComponent(
                name=f"button{i}", content=f"<button>Button {i}</button>"
            ).create(templates_dir)

        start_time = time.time()

        components.discover_components()

        end_time = time.time()
        scan_duration = end_time - start_time

        assert len(components._components) == 100
        assert scan_duration < 1.0

    def test_deep_directory_structure(self, templates_dir):
        current_dir = templates_dir
        depth = 10

        for i in range(depth):
            current_dir = current_dir / f"level{i}"
            current_dir.mkdir()
            TestComponent(
                name=f"button{i}",
                content=f"<button>Nested level {i}</button>",
                sub_dir="/".join(f"level{j}" for j in range(i + 1)),
            ).create(templates_dir)

        start_time = time.time()

        components.discover_components()

        end_time = time.time()
        scan_duration = end_time - start_time

        assert len(components._components) == depth
        assert scan_duration < 1.0

    def test_repeated_access_performance(self, templates_dir):
        TestComponent(name="button", content="<button>Performance</button>").create(
            templates_dir
        )

        # uncached
        start_time = time.time()
        components.get_component("button")
        first_access = time.time() - start_time

        # should be cached
        start_time = time.time()
        components.get_component("button")
        cached_access = time.time() - start_time

        # Cached access should be significantly faster
        assert cached_access < first_access / 2



from __future__ import annotations

import pytest

from django_bird.params import Param
from django_bird.params import Params
from django_bird.params import Value
from django_bird.templatetags.tags.prop import PropNode


class TestValue:
    @pytest.mark.parametrize(
        "value,context,expected",
        [
            (Value("btn", quoted=True), {"btn": "blue"}, "btn"),
            (Value("btn", quoted=False), {"btn": "blue"}, "blue"),
            (Value(True), {}, True),
            (Value(None), {}, None),
            (Value("True"), {}, True),
            (Value("False"), {}, None),
            (Value("undefined", quoted=False), {}, "undefined"),
            (Value("foo.bar", quoted=False), {}, "foo.bar"),
            (Value("foo.bar", quoted=False), {"foo": {}}, "foo.bar"),
        ],
    )
    def test_resolve(self, value, context, expected):
        assert value.resolve(context) == expected


class TestParam:
    @pytest.mark.parametrize(
        "param,context,expected",
        [
            (Param(name="class", value=Value("btn", quoted=False)), {}, 'class="btn"'),
            (
                Param(name="class", value=Value("btn", quoted=False)),
                {"btn": "blue"},
                'class="blue"',
            ),
            (Param(name="disabled", value=Value(True)), {}, "disabled"),
            (Param(name="disabled", value=Value(None)), {}, ""),
            (
                Param(name="class", value=Value("btn", quoted=True)),
                {"btn": "blue"},
                'class="btn"',
            ),
            (
                Param(name="class", value=Value("item.name", quoted=True)),
                {"item": {"name": "value"}},
                'class="item.name"',
            ),
            (
                Param(name="class", value=Value("undefined", quoted=False)),
                {},
                'class="undefined"',
            ),
            (
                Param(name="class", value=Value("foo.bar", quoted=False)),
                {},
                'class="foo.bar"',
            ),
            (
                Param(name="class", value=Value("foo.bar", quoted=False)),
                {"foo": {}},
                'class="foo.bar"',
            ),
            (Param(name="hx_get", value=Value("/", quoted=False)), {}, 'hx-get="/"'),
        ],
    )
    def test_render_attr(self, param, context, expected):
        assert param.render_attr(context) == expected

    @pytest.mark.parametrize(
        "param,context,expected",
        [
            (Param(name="class", value=Value("btn", quoted=False)), {}, "btn"),
            (
                Param(name="class", value=Value("btn", quoted=False)),
                {"btn": "blue"},
                "blue",
            ),
            (Param(name="disabled", value=Value(True)), {}, True),
            (Param(name="disabled", value=Value(None)), {}, None),
            (
                Param(name="count", value=Value("items.length", quoted=False)),
                {"items": {"length": 5}},
                5,
            ),
            (Param(name="isActive", value=Value("False")), {}, None),
            (Param(name="isActive", value=Value("True")), {}, True),
            (
                Param(name="data", value=Value("user.name", quoted=False)),
                {"user": {"name": "Alice"}},
                "Alice",
            ),
            (
                Param(name="data", value=Value("user.name", quoted=True)),
                {"user": {"name": "Alice"}},
                "user.name",
            ),
            (
                Param(name="class", value=Value("undefined", quoted=False)),
                {},
                "undefined",
            ),
            (
                Param(name="data", value=Value("user.name", quoted=False)),
                {},
                "user.name",
            ),
            (
                Param(name="data", value=Value("user.name", quoted=False)),
                {"user": {}},
                "user.name",
            ),
        ],
    )
    def test_render_prop(self, param, context, expected):
        assert param.render_prop(context) == expected

    @pytest.mark.parametrize(
        "bit,expected",
        [
            ("class='btn'", Param(name="class", value=Value("btn", quoted=True))),
            ('class="btn"', Param(name="class", value=Value("btn", quoted=True))),
            ("class=btn", Param(name="class", value=Value("btn", quoted=False))),
            ("disabled", Param(name="disabled", value=Value(True))),
            (
                "class=item.name",
                Param(name="class", value=Value("item.name", quoted=False)),
            ),
            (
                'class="item.name"',
                Param(name="class", value=Value("item.name", quoted=True)),
            ),
        ],
    )
    def test_from_bit(self, bit, expected):
        assert Param.from_bit(bit) == expected


class TestParams:
    @pytest.mark.parametrize(
        "params,nodelist,context,expected_props,expected_attrs",
        [
            (
                Params(attrs=[Param(name="class", value=Value(None))]),
                [PropNode(name="class", default="btn", attrs=[])],
                {},
                {"class": "btn"},
                [],
            ),
            (
                Params(attrs=[Param(name="class", value=Value("btn", quoted=False))]),
                [PropNode(name="class", default=None, attrs=[])],
                {},
                {"class": "btn"},
                [],
            ),
            (
                Params(attrs=[Param(name="class", value=Value("btn", quoted=False))]),
                None,
                {},
                None,
                [Param(name="class", value=Value("btn", quoted=False))],
            ),
            (
                Params(attrs=[Param(name="class", value=Value("static", quoted=True))]),
                [PropNode(name="class", default=None, attrs=[])],
                {"static": "dynamic"},
                {"class": "static"},
                [],
            ),
            (
                Params(attrs=[Param(name="class", value=Value("var", quoted=False))]),
                [PropNode(name="class", default=None, attrs=[])],
                {"var": "dynamic"},
                {"class": "dynamic"},
                [],
            ),
            (
                Params(
                    attrs=[Param(name="class", value=Value("undefined", quoted=False))]
                ),
                [PropNode(name="class", default=None, attrs=[])],
                {},
                {"class": "undefined"},
                [],
            ),
            (
                Params(
                    attrs=[Param(name="class", value=Value("user.name", quoted=False))]
                ),
                [PropNode(name="class", default=None, attrs=[])],
                {},
                {"class": "user.name"},
                [],
            ),
        ],
    )
    def test_render_props(
        self, params, nodelist, context, expected_props, expected_attrs
    ):
        assert params.render_props(nodelist, context) == expected_props
        assert params.attrs == expected_attrs

    @pytest.mark.parametrize(
        "params,context,expected",
        [
            (
                Params(attrs=[Param(name="class", value=Value("btn", quoted=False))]),
                {},
                'class="btn"',
            ),
            (
                Params(attrs=[Param(name="class", value=Value("btn", quoted=False))]),
                {"btn": "blue"},
                'class="blue"',
            ),
            (Params(attrs=[Param(name="disabled", value=Value(True))]), {}, "disabled"),
            (
                Params(
                    attrs=[
                        Param(name="class", value=Value("btn", quoted=False)),
                        Param(name="disabled", value=Value(True)),
                    ]
                ),
                {},
                'class="btn" disabled',
            ),
            (
                Params(attrs=[Param(name="class", value=Value("static", quoted=True))]),
                {"static": "dynamic"},
                'class="static"',
            ),
            (
                Params(attrs=[Param(name="class", value=Value("var", quoted=False))]),
                {"var": "dynamic"},
                'class="dynamic"',
            ),
            (
                Params(
                    attrs=[Param(name="class", value=Value("undefined", quoted=False))]
                ),
                {},
                'class="undefined"',
            ),
            (
                Params(
                    attrs=[Param(name="class", value=Value("user.name", quoted=False))]
                ),
                {},
                'class="user.name"',
            ),
            (
                Params(
                    attrs=[Param(name="class", value=Value("user.name", quoted=False))]
                ),
                {"user": {}},
                'class="user.name"',
            ),
        ],
    )
    def test_render_attrs(self, params, context, expected):
        assert params.render_attrs(context) == expected

    @pytest.mark.parametrize(
        "attrs,expected",
        [
            (
                [Param(name="class", value=Value("btn", quoted=True))],
                Params(attrs=[Param(name="class", value=Value("btn", quoted=True))]),
            ),
            (
                [
                    Param(name="class", value=Value("btn", quoted=True)),
                    Param(name="id", value=Value("my-btn", quoted=True)),
                ],
                Params(
                    attrs=[
                        Param(name="class", value=Value("btn", quoted=True)),
                        Param(name="id", value=Value("my-btn", quoted=True)),
                    ]
                ),
            ),
            (
                [Param(name="disabled", value=Value(True))],
                Params(attrs=[Param(name="disabled", value=Value(True))]),
            ),
            (
                [Param(name="class", value=Value("dynamic", quoted=False))],
                Params(
                    attrs=[Param(name="class", value=Value("dynamic", quoted=False))]
                ),
            ),
            (
                [
                    Param(name="class", value=Value("item.name", quoted=False)),
                    Param(name="id", value=Value("user.id", quoted=False)),
                ],
                Params(
                    attrs=[
                        Param(name="class", value=Value("item.name", quoted=False)),
                        Param(name="id", value=Value("user.id", quoted=False)),
                    ]
                ),
            ),
        ],
    )
    def test_with_attrs(self, attrs, expected):
        assert Params.with_attrs(attrs) == expected



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from django.contrib.staticfiles import finders
from django.urls import reverse

from ._typing import override


class AssetType(Enum):
    CSS = "css"
    JS = "js"

    @property
    def content_type(self):
        match self:
            case AssetType.CSS:
                return "text/css"
            case AssetType.JS:
                return "application/javascript"

    @property
    def ext(self):
        return f".{self.value}"

    @classmethod
    def from_tag_name(cls, tag_name: str):
        try:
            asset_type = tag_name.split(":")[1]
            match asset_type:
                case "css":
                    return cls.CSS
                case "js":
                    return cls.JS
                case _:
                    raise ValueError(f"Unknown asset type: {asset_type}")
        except IndexError as e:
            raise ValueError(f"Invalid tag name: {tag_name}") from e


@dataclass(frozen=True, slots=True)
class Asset:
    path: Path
    type: AssetType

    @override
    def __hash__(self) -> int:
        return hash((str(self.path), self.type))

    def exists(self) -> bool:
        return self.path.exists()

    def render(self):
        match self.type:
            case AssetType.CSS:
                return f'<link rel="stylesheet" href="{self.url}">'
            case AssetType.JS:
                return f'<script src="{self.url}"></script>'

    @property
    def url(self) -> str:
        component_name = self.path.stem
        asset_filename = self.path.name

        path = finders.find(f"django_bird/assets/{component_name}/{asset_filename}")

        return path or reverse(
            "django_bird:asset",
            kwargs={
                "component_name": component_name,
                "asset_filename": asset_filename,
            },
        )

    @classmethod
    def from_path(cls, path: Path) -> Asset:
        match path.suffix.lower():
            case ".css":
                asset_type = AssetType.CSS
            case ".js":
                asset_type = AssetType.JS
            case _:
                raise ValueError(f"Unknown asset type for path: {path}")
        return cls(path=path, type=asset_type)
